"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.raceSync = exports.race = exports.anySync = exports.any = exports.allSettledSync = exports.allSettled = exports.graphRunSync = exports.graphRun = exports.RunnerSync = exports.Runner = exports.RunnerBase = void 0;
const node_events_1 = require("node:events");
class RunnerBase {
    results = new Map();
    settled = new Map();
    dependents = new Map();
    directDependents = new Map();
    options;
    abortController;
    failFast;
    errors = [];
    from;
    constructor(options, from) {
        const ac = new AbortController();
        this.from = from ?? this.constructor;
        this.abortController = ac;
        (0, node_events_1.setMaxListeners)(Infinity, ac.signal);
        this.options = options;
        if (!options.graph.length) {
            const er = new Error('no nodes provided to graph traversal', {
                cause: {
                    found: options.graph,
                    wanted: '[first: Node, ...rest: Node[]]',
                },
            });
            Error.captureStackTrace(er, from);
            throw er;
        }
        this.failFast = options.failFast !== false;
        const { signal } = options;
        if (signal !== undefined) {
            signal.addEventListener('abort', reason => ac.abort(reason), {
                once: true,
                signal: ac.signal,
            });
        }
    }
    route(n, d) {
        const dependents = this.dependents.get(d);
        if (!dependents?.has(n))
            return undefined;
        const directDependents = this.directDependents.get(d);
        if (!directDependents)
            return undefined;
        if (directDependents.has(n)) {
            return [n, d];
        }
        const queue = [
            ...directDependents,
        ].map(dd => [dd, d]);
        let step = undefined;
        while (undefined !== (step = queue.shift())) {
            if (!dependents.has(step[0]))
                continue;
            if (step[0] === n) {
                return step;
            }
            const ddd = this.directDependents.get(step[0]);
            if (ddd) {
                for (const d of ddd) {
                    queue.push([d, ...step]);
                }
            }
        }
    }
    cycleCheck(n, path, d) {
        const dependents = this.dependents.get(n) ?? new Set();
        this.dependents.set(n, dependents);
        const isCycle = dependents.has(d);
        if (isCycle) {
            const cycle = this.route(d, n);
            if (!cycle) {
                throw new Error('cycle detected, but cycle route not found');
            }
            cycle.unshift(n);
            this.onCycle(d, cycle, path);
            return true;
        }
        const depDD = this.directDependents.get(d) ?? new Set();
        this.directDependents.set(d, depDD);
        depDD.add(n);
        const depDependents = this.dependents.get(d) ?? new Set();
        this.dependents.set(d, depDependents);
        for (const n of dependents) {
            depDependents.add(n);
        }
        depDependents.add(n);
        return false;
    }
    handleError(er, n, path) {
        this.errors.push(er);
        this.settled.set(n, {
            status: 'rejected',
            reason: er,
        });
        if (this.failFast) {
            this.abortController.abort(er);
            const e = new Error('failed graph traversal', {
                cause: {
                    node: n,
                    path,
                    cause: er,
                },
            });
            Error.captureStackTrace(e, this.from);
            throw e;
        }
    }
    handleValue(value, n) {
        this.results.set(n, value);
        this.settled.set(n, {
            status: 'fulfilled',
            value,
        });
    }
}
exports.RunnerBase = RunnerBase;
class Runner extends RunnerBase {
    running = new Map();
    async getDeps(n) {
        if (this.abortController.signal.aborted)
            return [];
        const deps = await this.options.getDeps(n);
        for (const d of deps) {
            const dependents = this.dependents.get(d) ?? new Set();
            this.dependents.set(d, dependents);
            dependents.add(n);
            const depDD = this.directDependents.get(d) ?? new Set();
            this.directDependents.set(d, depDD);
            depDD.add(n);
        }
        return deps;
    }
    async visit(n, path, depResults) {
        const { signal } = this.abortController;
        return this.options.visit(n, signal, path, depResults);
    }
    async onCycle(n, cycle, path) {
        if (this.abortController.signal.aborted)
            return;
        await this.options.onCycle?.(n, cycle, path);
    }
    async #walk(n, path) {
        const r = this.running.get(n);
        if (r)
            return r;
        if (this.settled.get(n))
            return;
        const p = this.#step(n, path).then(() => {
            this.running.delete(n);
        }, er => {
            this.running.delete(n);
            throw er;
        });
        this.running.set(n, p);
        return p;
    }
    async #step(n, path) {
        const dependents = this.dependents.get(n) ?? new Set();
        this.dependents.set(n, dependents);
        const deps = await this.getDeps(n);
        const awaiting = [];
        const depPath = [...path, n];
        for (const d of deps) {
            if (this.abortController.signal.aborted)
                return;
            if (d === n)
                continue;
            if (this.cycleCheck(n, depPath, d))
                continue;
            if (this.settled.get(d))
                continue;
            awaiting.push(this.running.get(d) ?? this.#walk(d, depPath));
        }
        if (this.abortController.signal.aborted)
            return;
        await Promise.all(awaiting);
        if (this.abortController.signal.aborted)
            return;
        const depRes = new Map(deps.map(d => [d, this.results.get(d)]));
        try {
            this.handleValue(await this.visit(n, path, depRes), n);
        }
        catch (er) {
            this.handleError(er, n, path);
        }
    }
    async run() {
        const promises = [];
        for (const n of this.options.graph) {
            promises.push(this.#walk(n, []));
        }
        await Promise.all(promises);
    }
}
exports.Runner = Runner;
class RunnerSync extends RunnerBase {
    getDeps(n) {
        if (this.abortController.signal.aborted)
            return [];
        return this.options.getDeps(n);
    }
    visit(n, path, depResults) {
        const { signal } = this.abortController;
        return this.options.visit(n, signal, path, depResults);
    }
    onCycle(n, cycle, path) {
        if (this.abortController.signal.aborted)
            return;
        this.options.onCycle?.(n, cycle, path);
    }
    #walk(n, path) {
        if (this.settled.get(n))
            return;
        this.#step(n, path);
    }
    #step(n, path) {
        const dependents = this.dependents.get(n) ?? new Set();
        this.dependents.set(n, dependents);
        const deps = this.getDeps(n);
        const depPath = [...path, n];
        for (const d of deps) {
            if (this.abortController.signal.aborted)
                return;
            if (d === n)
                continue;
            if (this.cycleCheck(n, depPath, d))
                continue;
            if (!this.settled.get(d))
                this.#walk(d, depPath);
        }
        if (this.abortController.signal.aborted)
            return;
        const depRes = new Map(deps.map(d => [d, this.results.get(d)]));
        try {
            this.handleValue(this.visit(n, path, depRes), n);
        }
        catch (er) {
            this.handleError(er, n, path);
        }
    }
    run() {
        for (const n of this.options.graph) {
            this.#walk(n, []);
        }
        return this.results;
    }
}
exports.RunnerSync = RunnerSync;
const graphRun = async (options) => {
    const runner = new Runner(options, exports.graphRun);
    await runner.run();
    if (runner.errors.length) {
        const e = new AggregateError(runner.errors, 'failed graph traversal');
        Error.captureStackTrace(e, exports.graphRun);
        throw e;
    }
    return runner.results;
};
exports.graphRun = graphRun;
const graphRunSync = (options) => {
    const runner = new RunnerSync(options, exports.graphRunSync);
    runner.run();
    if (runner.errors.length) {
        const e = new AggregateError(runner.errors, 'failed graph traversal');
        Error.captureStackTrace(e, exports.graphRunSync);
        throw e;
    }
    return runner.results;
};
exports.graphRunSync = graphRunSync;
const allSettled = async (options) => {
    const runner = new Runner({ ...options, failFast: false }, exports.allSettled);
    await runner.run();
    return runner.settled;
};
exports.allSettled = allSettled;
const allSettledSync = (options) => {
    const runner = new RunnerSync({ ...options, failFast: false }, exports.allSettledSync);
    runner.run();
    return runner.settled;
};
exports.allSettledSync = allSettledSync;
const any = async (options) => {
    const ac = new AbortController();
    let result;
    let found = false;
    const runner = new Runner({
        ...options,
        failFast: false,
        signal: ac.signal,
        visit: async (node, signal, path, depResults) => {
            try {
                result = await options.visit(node, signal, path, depResults);
                found = true;
                ac.abort('found');
            }
            finally {
                return result;
            }
        },
    }, exports.any);
    await runner.run();
    if (!found) {
        const e = new AggregateError(runner.errors, 'failed graph traversal');
        Error.captureStackTrace(e, exports.any);
        throw e;
    }
    return result;
};
exports.any = any;
const anySync = (options) => {
    const ac = new AbortController();
    let result;
    let found = false;
    const runner = new RunnerSync({
        ...options,
        failFast: false,
        signal: ac.signal,
        visit: (node, signal, path, depResults) => {
            try {
                result = options.visit(node, signal, path, depResults);
                found = true;
                ac.abort('found');
            }
            finally {
                return result;
            }
        },
    }, exports.anySync);
    runner.run();
    if (!found) {
        const e = new AggregateError(runner.errors, 'failed graph traversal');
        Error.captureStackTrace(e, exports.anySync);
        throw e;
    }
    return result;
};
exports.anySync = anySync;
const race = async (options) => {
    const ac = new AbortController();
    let result;
    const runner = new Runner({
        ...options,
        failFast: true,
        signal: ac.signal,
        visit: async (node, signal, path, depResults) => {
            try {
                result = await options.visit(node, signal, path, depResults);
                ac.abort('found');
            }
            finally {
                return result;
            }
        },
    }, exports.race);
    await runner.run();
    return result;
};
exports.race = race;
const raceSync = (options) => {
    const ac = new AbortController();
    let result;
    const runner = new RunnerSync({
        ...options,
        failFast: true,
        signal: ac.signal,
        visit: (node, signal, path, depResults) => {
            try {
                result = options.visit(node, signal, path, depResults);
                ac.abort('found');
            }
            finally {
                return result;
            }
        },
    }, exports.race);
    runner.run();
    return result;
};
exports.raceSync = raceSync;
//# sourceMappingURL=index.js.map